/*

This is FUNCPACK.C, a Personal REXX function package example.
Coded for Microsoft C version 6.0 small, medium, or large models.

Function:

    Install REXX function package, then terminate and stay resident.
    Subsequent calls to the function package are handled.

Sample Usage:

    To install (from DOS prompt):

    FUNCPACK

    later, from a REXX program:

    say hello()

    or, run the FUNCDEMO.REX file.

Include Files:

    REXXCALL.H   - maps REXX control block

Calls:

    FINDINT.ASM  - gets address of REXX interrupt handler
    REXXFCN.ASM  - environment entry point, calls hndutil function.

Notes:

    To recreate executable version of this program use the command

        MAKE APIMSC                     (small model)
        MAKE MODEL=-AM APIMSC           (medium model)
        MAKE MODEL=-AL APIMSC           (large model)

    FUNCPACK.C uses C preprocessor statements to generate
    the correct code for any of the three models.

    FINDINT.ASM and REXXFCN.ASM also need no changes for the
    different models.

    Use LINK options or EXEMOD to reduce memory requested when
    FUNCPACK is loaded.

    This program was tested with Personal REXX 3.00 and MSC 6.0.

*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dos.h>
#include <malloc.h>

#include "rexxcall.h"

#define MAKE_LONG(hi, low) \
 ((long) ((((unsigned long) (unsigned) (hi)) << 16) + \
 (unsigned long) (unsigned) (low)))
#define HIGH_WORD(l) ((unsigned) (((unsigned long) (l)) >> 16))
#define LOW_WORD(l) ((unsigned) (((unsigned long) (l)) & 0xffff))

#ifdef M_I86SM  /* small model */
#define FINDINT findints
#endif
#ifdef M_I86MM  /* medium model */
#define FINDINT findintm
#endif
#ifdef M_I86LM  /* large model */
#define FINDINT findintl
#endif

/* funp type is pointer to function returning void */
typedef void (*funp)(void);
typedef struct
   {
   char *name;     /* ptr to name of a function */
   char argmin;    /* minimum number of allowable arguments */
   char argmax;    /* maximum number of allowable arguments */
   funp f;         /* ptr to handler for this function */
   } table_entry;

extern int rexxfcn();

/* function prototypes */
int FINDINT (int *);
int hndutil (struct call_block far *);
funp which (char *);
unsigned showsp (void);
int has_null (char *, int);
int args_have_null (void);
void exit_tsr (void);
void fixpsp (void);
void nomem (void);
void error (char *);

static char return_buffer[256];
static int result, return_length, nargs;
static unsigned data_segment;
static char **arg;
static int *length;

union REGS inregs, outregs;
struct SREGS segregs;

void main (int argc, char **argv)
    {
    struct env_block eb, *ebp = &eb;

    int rexx_interrupt, interrupt_status;

    rexx_interrupt = FINDINT(&interrupt_status);
    switch (interrupt_status)
       {
    case 1:
       error("RXINT environment setting invalid");
       break;
    case 2:
       error("REXX interrupt in use by another program");
       break;
    case 3:
       error("REXX Interrupt Manager not loaded\nUse RXINTMGR command to load it");
       break;
       }

    /* setup environment block telling rexx who and where we are */
    memset(&eb, 0, sizeof(struct env_block));
    memcpy(eb.env_name, "SAMPLE  ", 8);
    segread(&segregs);
    eb.env_entry_addr = (FARP) rexxfcn; /* assembler gate for calls from rexx */

#ifdef M_I86SM   /* predefined by microsoft c if small model */
    /* 0 for segment address of hndutil will indicate to rexxfcn */
    /* to use near call to hndutil */
    eb.env_exec_addr = (FARP) MAKE_LONG(0, hndutil);  /* function handler */
#else            /* medium or large model */
    eb.env_exec_addr = (FARP) hndutil;          /* function handler */
#endif

    /* setup segment register values to be put into effect by rexxfcn when */
    /* our package is called */
    eb.env_dseg = data_segment = segregs.ds;    /* our data segment */
    eb.env_sseg = segregs.ss;                   /* our stack segment */
    eb.env_sp = showsp();

    /* announce ourselves to rexx */
    inregs.h.ah = REXX_FCN;
    inregs.h.al = FCN_INSTALL;
#ifdef M_I86LM    /* predefined by microsoft c if large model */
    inregs.x.bx = FP_SEG(ebp);
    inregs.x.dx = FP_OFF(ebp);
#else       /* medium or small model */
    inregs.x.bx = data_segment;
    inregs.x.dx = (unsigned) &eb;
#endif
    int86x(rexx_interrupt, &inregs, &outregs, &segregs);

    if (outregs.x.ax != 0)
       {
       printf("Error %d installing function package.\n",outregs.x.ax);
       exit(2);
       }

    /* terminate and stay resident, awaiting calls from rexx to hndutil */
    exit_tsr();
    }

/* handle function requests */
int hndutil (struct call_block far *cb)
    {
    char *fcn;
    funp f;
    struct arg_block far *argp;
    int i, l;

    result = 0;

    /* note the number of arguments passed */
    nargs = (int)cb->arg_count;

    /* if name length is 0, this is a termination call from REXX */
    /* this example does nothing special for termination */
    if (cb->name_length == 0)
        return 1;

    /* create a string containing the name of the function involved */
    fcn = malloc(cb->name_length + 1);
    if (fcn == NULL)
        nomem();
#ifdef M_I86LM   /* large model */
    memcpy(fcn, cb->name_addr, cb->name_length);
#else      /* small and medium models */
    _fmemcpy(fcn, cb->name_addr, cb->name_length);
#endif
    fcn[cb->name_length] = '\0';

    /* see if we can handle the function */
    f = which(fcn);
    free(fcn);
    if (f == (funp) NULL)
        return result;

    /* for easier access to arguments, copy then from rexx's memory  */
    /* into our data segment.                   */
    /* null ('\0') character is appended to arguments so they can   */
    /* be accessed with c string handling functions when appropriate */
    /* it is not always appropriate, since rexx strings, unlike c    */
    /* strings, can contain embedded null characters.          */
    /* functions has_null and args_have_null are provided to test   */
    /* whether any of the arguments contain null characters       */
    arg = (char **) malloc(nargs*sizeof(char *));
    if (arg == NULL)
        nomem();
    length = (int *) malloc(nargs*sizeof(int));
    if (length == NULL)
        nomem;
    argp = (struct arg_block far *) cb->arg_list_addr;
    for (i=0; i<nargs; i++, argp++)
        {
        if (argp->arg_addr != (FARP) NULL)
            {
            arg[i] = malloc((length[i] = (int)argp->arg_length) + 1);
            if (arg[i] == NULL)
                nomem();
#ifdef M_I86LM      /* large model */
            memcpy(arg[i], argp->arg_addr, length[i]);
#else               /* small and medium models */
            _fmemcpy(arg[i], argp->arg_addr, length[i]);
#endif
            l = length[i];
            arg[i][l] = '\0';
            }
        else
            {
            /* allow for omitted arguments */
            arg[i] = NULL;
            length[i] = 0;
            }
       }

    /* call our routine to handle the function */
    (*f)();

    /* tell rexx address and length of the returned value */
    cb->ret_count = return_length;
    cb->ret_addr = (FARP) MAKE_LONG(data_segment, (unsigned) &return_buffer[0]);

    /* release temporary memory */
    for (i=0; i<nargs; i++)
        if (arg[i] != NULL)
            free(arg[i]);
    free(arg);
    free(length);

    fixpsp();

    return result;
    }

/* decide whether the function is handled in this package  */
/* returns a pointer to routine to handle it, or NULL if   */
/* we can't handle it,                                     */
/* with result=1 if it's not one of our functions          */
/*  result=2 if too many or too few arguments passed   */
funp which (char *fcn)
    {
    void hello();
    int l, i;
    static table_entry table[] =
       {
            { "HELLO", 0, 5, hello },
            /* add other functions here */
            { NULL }
       };

    l = strlen(fcn);
    if (!l)
        return NULL;
    printf("which: %s\n", fcn);
    for (i=0; table[i].name!=NULL; i++)
        {
        if (strcmpi(fcn, table[i].name) != 0)
            continue;
        if (nargs<table[i].argmin || nargs>table[i].argmax)
            {
            result = 2;
            return NULL;
            }
        return table[i].f;
       }
    result = 1;
    return NULL;
    }

#ifdef M_I86LM
/* returns the current stack pointer */
unsigned showsp (void)
    {
    int dummy, *dummyp = &dummy;

    return FP_OFF(dummyp);
    }
#else
/* returns the current stack pointer */
unsigned showsp (void)
    {
    int dummy; /* dummy variable allocated on stack so we can get its address */

    return (unsigned) &dummy;
    }
#endif

/* returns 1 if a 'string' of length l contains any null characters */
int has_null (char *s, int l)
    {
    return memchr(s, '\0', l) != NULL;
    }

/* returns 1 if any of the arguments contains a null character */
int args_have_null (void)
    {
    int i;

    for (i = 0; i < nargs; i++)
       if (length[i] > 0 && has_null(arg[i], length[i]))
        return 1;

    return 0;
    }

/* terminate and stay resident */
void exit_tsr (void)
    {
    unsigned far *memp;
    int i;

    /* close stdxxx files */
    for (i=0; i<5; i++) {
        inregs.x.ax = 0x3e00;
        inregs.x.bx = i;
        int86(0x21, &inregs, &outregs);
        }

    /* point at second word of our psp, which contains our ending paragraph */
    memp = (unsigned far *) MAKE_LONG(_psp, 2);

    inregs.x.ax = 0x3100; /* dos terminate and stay resident function  */
    inregs.x.dx = *memp - _psp;  /* paragraphs of memory allocated to us */

    int86(0x21, &inregs, &outregs);

    /* we never return here */
    }

/* fixup memory allocation block                 */
/* microsoft c's memory allocation routines occasionally call dos to       */
/* allocate additional memory. in our case (a terminate-stay-resident pgm) */
/* this memory allocation will fail. unfortunately, the dos code to    */
/* handle this call marks our memory as owned by the currently-executing   */
/* program (usually rx), and when the rexx program finishes, this causes   */
/* our memory to be freed                     */
/* so before we return from our function handler, we always re-mark our    */
/* memory as owned by us, to prevent its being freed by dos        */
/* we do this by putting our psp address into the dos memory header block  */
/* located one paragraph before our psp                */
void fixpsp (void)
    {
    unsigned far *memp;

    memp = (unsigned far *) MAKE_LONG(_psp - 1, 1);
    *memp = _psp;
    }

/* a sample function */
/* it prints out its arguments and returns "goodbye" to rexx */
void hello (void)
    {
    int i;

    printf("hello\n");
    printf("%d argument(s) passed to hello function\n", nargs);
    for (i = 0; i < nargs; i++)
        {
        if (arg[i] == NULL)
            printf("argument %d is omitted\n", i + 1);
        else if (length[i] == 0)
            printf("argument %d is the null string\n", i + 1);
        else if (has_null(arg[i], length[i]))
            printf("argument %d contains a null character\n", i + 1);
        else
            printf("argument %d = %s\n", i + 1, arg[i]);
        }
    strcpy(return_buffer, "goodbye");
    return_length = 7;
    }

/* handle out to memory condition */
/* this should be replaced with something more forgiving */
void nomem (void)
    {
    error("FUNCPACK: insufficient memory");
    }

/* display error message and exit */
void error (char *s)
    {
    printf("ABORT:- %s\n",s);
    exit(255);
    }
